#include "graphplan.h"

inline bool PlanGraph::checkInterference(const OpNode& a, const OpNode& b) {
	//Two nops will never interfere
	if (!a.ptr && !b.ptr) return 0;

	//Nops vs. regular operations are simple (only one fact)
	else if (!a.ptr) return b.ptr->del.elem( const_cast<Predicate*&>(predNodes[a.pre.front()].ptr));
	else if (!b.ptr) return a.ptr->del.elem( const_cast<Predicate*&>(predNodes[b.pre.front()].ptr));
	
	//Compare delete list of a against add and precondition lists of b
	for (set<Predicate*>::const_iterator del = a.ptr->del.preds.begin(); del!=a.ptr->del.preds.end(); del++) {
		if (b.ptr->add.preds.find(*del)!=b.ptr->add.preds.end()) return 1;
		if (b.ptr->pre.preds.find(*del)!=b.ptr->pre.preds.end()) return 1;
	}
	//And the other way round
	for (set<Predicate*>::const_iterator del = b.ptr->del.preds.begin(); del!=b.ptr->del.preds.end(); del++) {
		if (a.ptr->add.preds.find(*del)!=a.ptr->add.preds.end()) return 1;
		if (a.ptr->pre.preds.find(*del)!=a.ptr->pre.preds.end()) return 1;
	}
	return 0;
}

//Iterate over all facts and check for exclusion
inline bool PlanGraph::addOpIfPossible(const Operation& o, int i) {
	for (size_t op=0; op<opLayers[i]; op++) {if (opNodes[op].ptr==&o) return 0;}

	vector<size_t> checked;
	for (size_t p=0; p<predLayers[i]; p++) {
		if (o.pre.elem(const_cast<Predicate*>(predNodes[p].ptr))) {
			for (size_t prev=0; prev<checked.size(); prev++) 
				if (predNodes[p].excl[i-1].get(checked[prev])) return 0;
			checked.push_back(p);
		}
	}
	//Couldn't find all predicates
	if (checked.size()<o.pre.preds.size()) return 0;
	//Create node and hook up with preconditions
	size_t on = newOp(&o);
	for (size_t pre=0; pre<checked.size(); pre++) opNodes[on].pre.push_back(checked[pre]);
	return 1;
}

void PlanGraph::buildLayer(int i) {
	if (!changed) return;
	changed=0;
	//Need to copy all predicate nodes of the last layer manually to avoid copying all the edges as well
	//Correction: I'm an idiot for having all the redundant nodes and edges in the first place,
	//so only do the nops for predicates that were introduced in the previous layer
	for (size_t p=predLayers[i-1]; p<predLayers[i]; p++) {
		size_t on = newOp(0);
		opNodes[on].add.push_back(p);
		opNodes[on].pre.push_back(p);
		predNodes[p].add.insert(predNodes[p].add.begin(), on);
	}
	
	//Changed to only search operation instances that haven't been used already
	//Thought about adding a list for self excluding operations, so we don't have to check them all
	//when no new facts showed up anyway
	for (list<Operation*>::iterator op = unusedOps.begin(); op!=unusedOps.end();) {
		if (!addOpIfPossible(**op++, i)) continue;
		list<Operation*>::iterator tmp=--op;
		size_t on = opNodes.size()-1; //A node was created and pushed to the back

		//Insert new predicates and connect with add edges to next layer
		//When inserting a new predicate node, check if we already have a node for this predicate.
		for (set<Predicate*>::iterator pred=(*op)->add.preds.begin(); pred!=(*op)->add.preds.end(); pred++) {
			size_t p=0;
			for (; p<predNodes.size(); p++) if (predNodes[p].ptr==*pred) break;
			if (p==predNodes.size()) p=newPred(*pred);
			opNodes[on].add.push_back(p);
			predNodes[p].add.push_back(on);
		}
		//The same for del edges to... uhm... graphplan doesn't really care about them)
		op++;
		unusedOps.erase(tmp);
	}
	finishLayer();
	addOpExclusions(i);
	addPredExclusions(i);
}

//These two would benefit quite a lot from bit vectors, but so far they aren't using too much time
//and bit vectors would require extra work in other places
inline bool PlanGraph::checkPredExclusion(const PredNode& a, const PredNode& b, const size_t& i) {
	for (vector<size_t>::const_iterator it=a.add.begin(); it!=a.add.end(); it++) {
		for (vector<size_t>::const_iterator it2=b.add.begin(); it2!=b.add.end(); it2++)
			if (!opNodes[*it].excl[i].get(*it2)) return 0;
	}
	return 1;
}

inline bool PlanGraph::checkOpExclusion(const OpNode& a, const OpNode& b, const size_t& i) {
	for (vector<size_t>::const_iterator it=a.pre.begin(); it!=a.pre.end(); it++) {
		for (vector<size_t>::const_iterator it2=b.pre.begin(); it2!=b.pre.end(); it2++)
			if (predNodes[*it].excl[i-1].get(*it2)) return 1;
	}
	return 0;
}

//Same as for operations, sans interference, plus registering if exclusions were updated
inline void PlanGraph::addPredExclusions(int i) {
	for (size_t p1 = 0; p1 < predLayers[i]; p1++) {
		predNodes[p1].excl.resize(i+1);
		predNodes[p1].excl[i].resize(predLayers[i+1]);
		for (size_t p2=0; p2<p1; p2++) {
			if (predNodes[p2].excl[i-1].get(p1)) {
				if (checkPredExclusion(predNodes[p1], predNodes[p2], i)) {
					predNodes[p1].excl[i].set(p2);
					predNodes[p2].excl[i].set(p1);
				}
				else changed=1;
			}
		}
	}
	for (size_t p1 = predLayers[i]; p1 < predLayers[i+1]; p1++) {
		PredNode& p = predNodes[p1];
		p.excl.resize(i+1);
		p.excl[i].resize(predLayers[i+1]);
		for (size_t p2=0; p2<p1; p2++) {
			if (checkPredExclusion(predNodes[p1], predNodes[p2], i)) {
				predNodes[p1].excl[i].set(p2);
				predNodes[p2].excl[i].set(p1);
			}
		}			
	}
}

//Might look redundant to have two bit vectors for exclusion and interference, when exclusion
//is just the combination of interference and competing needs, but it avoids lots of rechecking
//and is useful to display the exclusion matrix.
inline void PlanGraph::addOpExclusions(int i) {
	//Old nodes only have to recheck if they were mean last time
	for (size_t op1 = 0; op1<opLayers[i]; op1++) {
		opNodes[op1].interference.resize(opLayers[i+1]);
		opNodes[op1].excl.push_back(opNodes[op1].interference);
		for (size_t op2 = 0; op2<op1; op2++) {
			if (!opNodes[op2].interference.get(op1) && opNodes[op2].excl[i-1].get(op1) &&  
										checkOpExclusion(opNodes[op1], opNodes[op2], i)) {
				opNodes[op2].excl[i].set(op1);
				opNodes[op1].excl[i].set(op2);
			}
		}
	}
	//New nodes get the full enchilada, including interference
	for (size_t op1 = opLayers[i]; op1<opLayers[i+1]; op1++) {
		opNodes[op1].excl.resize(i+1);
		opNodes[op1].excl[i].resize(opLayers[i+1]);
		opNodes[op1].interference.resize(opLayers[i+1]);

		for (size_t op2 = 0; op2<op1; op2++) {
			if (checkInterference(opNodes[op1], opNodes[op2])) {
				opNodes[op1].interference.set(op2);
				opNodes[op2].interference.set(op1);
				opNodes[op1].excl[i].set(op2);
				opNodes[op2].excl[i].set(op1); 
			}
			else if (checkOpExclusion(opNodes[op1], opNodes[op2], i)) {
				opNodes[op1].excl[i].set(op2);
				opNodes[op2].excl[i].set(op1); 
			}
		}
	}
}
