#ifndef MAPME_INCLUDED
#define MAPME_INCLUDED

#include <string>
#include <map>
#include <vector>

#include "errors.h"

using std::string;
using std::map;
using std::vector;

//Forward declaration to make gcc happy (actually surprising VC++ didn't complain)
struct Atom;
//struct Constraint;

//Templates are your friend, helping to avoid all the annoying redundant code and copy/paste bugs
//Large parts of Atom, Predicate and Operation are the same but can't use inheritance
template<class T>
struct MapMe {
	string name;
	
	//storing all atoms and prototypes for operations and predicates
	static map<string, T*> Map;
	
	//We don't want dangling pointers when creating duplicates, so don't just
	//blindly overwrite the entry, but return the existing object
	static T* create(const char* n) {
		T*& p = Map[n];
		if (p) error(string(n)+" already exists\n");
		return p ? p : p=new T(n);
	}

	//Boring, return 0 if no entry was found, else a pointer to the object
	//Yes, this weird "typename" issue with gcc cost me about 30min of googling
	static T* getByName(const char* n) {
		typename map<string, T*>::iterator it = Map.find(n);
		return (it!=Map.end()) ? it->second : 0;
	}

	//Iterate over everything and let it print itself
	static void printAll() {
		for (typename map<string, T*>::iterator it=Map.begin(); it!=Map.end(); it++) {
			if (it!=Map.begin()) printf(", ");
			it->second->print();
		}
		printf("\n");
	}

	//Layout is slightly weird and complicated to avoid deleting twice or creating memory leaks
	//Just don't worry about this part, especially the garbage-collector-spoiled Java crowd
	static void deepWipe() {
		for (typename map<string, T*>::iterator it=Map.begin(); it!=Map.end(); it++) {
			vector<Atom*>::iterator pit = it->second->params.begin();
			for (; pit!=it->second->params.end(); pit++) delete *pit;
			delete it->second;
		}
		Map.clear();
	}

protected:
	//Nobody but derived classes are supposed to do this
	MapMe(const char* n) : name(n) {}
};

template<class T>
map<string, T*> MapMe<T>::Map;

#endif
