#include "operation.h"

vector<Operation> Operation::instances;
bool Operation::noDups=1, Operation::ignoreCons=0, Operation::filterRedOps=1;

//Addressing parameter variables by name is more comfortable and necessary for this whole console thing to work
Variable* Operation::getVariable(const char* n) {
	for (vector<Atom*>::iterator it=params.begin(); it!=params.end(); it++)
		if ((*it)->name==n) return dynamic_cast<Variable*>(*it);
	return 0;	
}

//Cleaning up. Move along, nothing to see here.
void Operation::wipe() {
	for (map<string, Operation*>::iterator it=Map.begin(); it!=Map.end(); it++) {
		set<Constraint*>::iterator cit=it->second->constraints.begin();
		for (; cit!=it->second->constraints.end(); cit++) delete *cit;
	}
	deepWipe();
	instances.clear();
}

string Operation::toString() const {
	string s=name+"(";
	for (vector<Atom*>::const_iterator it=params.begin(); it!=params.end(); it++) {
		if (it!=params.begin()) s+= isProto ? ", " : ",";
		if (isProto) s+=(*it)->type->name+":";
		s+=(*it)->name;
	}
	return s+=")\n";
}

int Operation::countInstances() const {
	int num=0;
	for (vector<Operation>::iterator it=instances.begin(); it!=instances.end(); it++) {
		if (it->name==name) num++;
	}
	return num;
}

void Operation::print() const {
	string head=toString();
	printf("\n%s - (%d instances)\n", head.substr(0, head.size()-1).c_str(), countInstances());
	printf("  pre: "); pre.print();
	printf("  del: "); del.print();
	printf("  add: "); add.print();

	printf("  const: ");
	for (set<Constraint*>::const_iterator it=constraints.begin(); it!=constraints.end(); it++) {
		if (it!=constraints.begin()) printf(" ^ ");
		printf((*it)->toString().c_str());
	}
	printf("\n");
}

void Operation::printAllInstances() {
	for (vector<Operation>::iterator it=instances.begin(); it!=instances.end(); it++)
		printf("%s", it->toString().c_str());
}

//Iterate over all predicates and parameters,
//change pointer froms variables to their assigned atoms
void replace(State& trg, State& src) {
	for (set<Predicate*>::iterator it=src.preds.begin(); it!=src.preds.end(); it++) {
		//Ugly: we copy, change, insert to avoid duplicates which in turn cause pesky issues with set::find
		//Just in case you were wondering about this stupid method
		Predicate tmp = **it;
		for (vector<Atom*>::iterator pit=tmp.params.begin(); pit!=tmp.params.end(); pit++) {
			Variable* v = dynamic_cast<Variable*>(*pit);
			if (v) *pit=v->value;
		}
		trg.preds.insert(Predicate::instance(tmp.name.c_str(), tmp.params));
	}
}

Operation* Operation::instance(const char* n, const vector<Atom*>& param) {
	Operation* p = getByName(n);
	if (!p) {
		error(string("Operation \"")+n+"\" doesn't exist\n");
		return 0;
	}
	if (param.size()!=p->params.size()) {
		error("Wrong number of parameters\n");
		return 0;
	}
	
	for (int i=0; i<(int)param.size(); i++) {
		//Instances can't include any Variables
		if (dynamic_cast<Variable*>(param[i])) {
			error("Sorry, no partially instanced operations allowed yet\n");
			return 0;
		}
		//Make sure the types all match the prototype
		if (p->params[i]->type!=anyType) {
			const Type* t=param[i]->type;
			do if (t==p->params[i]->type) break;
			while (t=t->parent);
			if (!t) {
				error("Type mismatch, "+param[i]->name+" is type "+param[i]->type->name+
							", but "+p->params[i]->type->name+" expected\n");
				return 0;
			}
		}
		
		//Make sure no atom is used as more than one parameter
		if (noDups)
			for (int j=i+1; j<(int)param.size(); j++) if (param[i]==param[j]) return 0;
		//Assign passed atoms to variables in prototype
		dynamic_cast<Variable*>(p->params[i])->value=param[i];
	}

	//Check all constraints if not set to ignore
	if (!ignoreCons) {
		for (set<Constraint*>::iterator it=p->constraints.begin(); it!=p->constraints.end(); it++) {
			if (!(*it)->check()) return 0;
		}
	}

	//Somewhat inefficient fly weight pattern to avoid a gazillion instances flying around
	//Note: there should never be duplicates created anyway, but what the hell...
	for (vector<Operation>::iterator it=instances.begin(); it!=instances.end(); it++)
		if (it->name==n && it->params==param) return &*it;

	//Create instance, copy effects from prototype
	instances.push_back(Operation(n));
	Operation& np = instances.back();
	np.params=param;
	np.isProto=0;

	//Replace all variables
	replace(np.add, p->add);
	replace(np.del, p->del);
	replace(np.pre, p->pre);

	//We check if this operation turned out to have the same preconditions and effects as a previous instance
	//If so, we delete it again and return the previous instance instead (not exactly efficient)
	//This drastically speeds up problems like the fridge example, where otherwise you would get lots of
	//operations where only the order of parameters varies, even when it doesn't matter
	if (filterRedOps) {
		for (vector<Operation>::iterator it=instances.begin(); it!=instances.end()-1; it++) {
			if (np.pre.preds==it->pre.preds && np.add.preds==it->add.preds && np.del.preds==it->del.preds) {
				instances.pop_back();
				return &*it;
			}
		}
	}
	
	return &np;
}

//Also giving fancy names to old stuff to look better on boxes and press releases is kind of "in", so:
//"I dub thee ye olde clockwork method, thou shalt be known as yet another bit of hot air"
void Operation::instanceOperations() {	
	instances.clear();
	
	map<string, Operation*>::iterator opit=Operation::Map.begin();
	for (; opit!=Operation::Map.end(); opit++) {
		Operation* op = opit->second;
		const int numParams = (int)op->params.size();
		vector<Atom*>* posAtoms = new vector<Atom*>[numParams];
		vector<Atom*>::iterator* paramIt = new vector<Atom*>::iterator[numParams];

		//Set everything to start, collect all allowed atoms for each variable and abort if there is none
		//Could be faster by storing atoms seperated by type in the first place
		//Hint: Yes, there is a perfectly legal use for goto's and breaking out of nested loops is one
		//Note: older versions of VC++ put int i in the wrong scope, allowing to break and repeat the test
		for (int i=0; i<numParams; i++) {
			for (map<string, Atom*>::iterator it = Atom::Map.begin(); it!=Atom::Map.end(); it++) {
				if (op->params[i]->type==anyType) posAtoms[i].push_back(it->second);
				else {
					const Type* t=it->second->type;
					do if (t==op->params[i]->type) break;
					while (t=t->parent);
					if (t) posAtoms[i].push_back(it->second);
				}
			}
			if (!posAtoms[i].size()) goto abort;
			paramIt[i]=posAtoms[i].begin();
		}

		for (;;) {
			//Build parameter vector from all current iterators
			vector<Atom*> par;
			for (int i=0; i<numParams; i++) par.push_back(*paramIt[i]);
			Operation::instance(op->name.c_str(), par);

			//The actual "clockwork". If last "digit" reaches end, reset and increase next in line
			//If the first digit would reset, we are finished
			int step=numParams;
			while (--step>=0) {
				paramIt[step]++;
				if (paramIt[step]==posAtoms[step].end()) paramIt[step]=posAtoms[step].begin();
				else break;

			}
			if (step<0) break; //First counter just wrapped, we are done
		}
abort:
		delete[] posAtoms;
		delete[] paramIt;
	}
}
